from unittest.case import TestCase
from unittest.mock import patch

from clinner.command import command
from clinner.run import HealthCheckMixin
from clinner.run.main import Main


class FooMain(HealthCheckMixin, Main):
    def health_check(self):
        return True


class HealthCheckMixinTestCase(TestCase):
    def setUp(self):
        self.cli_patcher = patch('clinner.run.base.CLI')
        self.cli_patcher.start()

    def test_main_add_arguments(self):
        @command
        def foo(*args, **kwargs):
            pass

        args = ['-r', '3', 'foo']
        main = FooMain(args)

        self.assertEqual(main.args.retry, 3)

    def test_main_run(self):
        @command
        def foo(*args, **kwargs):
            return 0

        args = ['-q', 'foo']
        main = FooMain(args)
        result = main.run()

        self.assertEqual(result, 0)

    def test_main_health_check_fails(self):
        @command
        def foo(*args, **kwargs):
            return 0

        args = ['-q', '-r', '1', 'foo']
        main = FooMain(args)
        main.health_check = lambda: False
        result = main.run()

        self.assertEqual(result, 1)

    def test_main_skip_health_check(self):
        @command
        def foo(*args, **kwargs):
            return 0

        args = ['-q', '--skip-check', 'foo']
        main = FooMain(args)
        main.health_check = lambda: False
        result = main.run()

        self.assertEqual(result, 0)

    def test_main_retry_zero(self):
        @command
        def foo(*args, **kwargs):
            return 0

        args = ['-q', '-r', '0', 'foo']
        main = FooMain(args)
        main.health_check = lambda: False
        result = main.run()

        self.assertEqual(result, 0)

    def tearDown(self):
        self.cli_patcher.stop()
        del command.register['foo']



class WrongCommandError(KeyError):
    pass


class WrongResourceError(KeyError):
    pass


class CommandTypeError(TypeError):
    pass


class CommandArgParseError(ValueError):
    pass


class ImproperlyConfigured(Exception):
    pass



# -*- coding: utf-8 -*-
"""Django application config module.
"""
try:
    from django.apps import AppConfig
except ImportError:
    AppConfig = object

from clinner.settings import settings


class Clinner(AppConfig):
    name = 'clinner'
    verbose_name = 'Clinner'

    def ready(self):
        settings.build_from_django()



import shlex

from clinner.command import Type, command

__all__ = ['dist']

VERSION_CHOICES = ('patch', 'minor', 'major')


@command(command_type=Type.SHELL,
         args=((('version',), {'help': 'Bump version', 'choices': VERSION_CHOICES}),),
         parser_opts={'help': 'Bump version, create package and upload it'})
def dist(*args, **kwargs):
    """
    Bump version, create package and upload it.
    """
    clean = shlex.split('rm -rf dist')
    bumpversion = shlex.split('bumpversion %s' % kwargs['version'])
    package = shlex.split('python setup.py sdist bdist_wheel')
    upload = shlex.split('twine upload dist/*')
    return [clean, bumpversion, package, upload]



import shlex

from clinner.command import Type, command

__all__ = ['nose']


@command(command_type=Type.SHELL,
         args=((('test_module',), {'nargs': '*', 'default': ['.'], 'help': 'Module to test'}),),
         parser_opts={'help': 'Run unit tests'})
def nose(*args, **kwargs):
    """
    Run unit tests using Nose.
    """
    coverage_erase = shlex.split('coverage erase')

    tests = shlex.split('nosetests')
    tests += args

    return [coverage_erase, tests]



import argparse
import os
from abc import ABCMeta, abstractmethod
from importlib import import_module
from subprocess import Popen

from clinner.builder import Builder
from clinner.cli import CLI
from clinner.command import Type, command
from clinner.exceptions import CommandArgParseError, CommandTypeError
from clinner.settings import settings

__all__ = ['MainMeta', 'BaseMain']


class MainMeta(ABCMeta):
    def __new__(mcs, name, bases, namespace):  # noqa
        def inject(self):
            """
            Add all environment variables defined in all inject methods.
            """
            for method in [v for k, v in namespace.items() if k.startswith('inject_')]:
                method(self)

        def add_arguments(self, parser, parser_class=None):
            """
            Add command line arguments to parser.

            :param parser: Parser.
            :param parser_class: Parser class.
            """
            self._commands_arguments(parser, parser_class)

            for base in reversed(bases):
                if hasattr(base, 'add_arguments'):
                    getattr(base, 'add_arguments')(self, parser)

            if hasattr(self, 'add_arguments'):
                self.add_arguments(parser)

        namespace['inject'] = inject
        namespace['_add_arguments'] = add_arguments

        for command_fqn in namespace.get('commands', []):
            try:
                m, c = command_fqn.rsplit('.', 1)
                if c not in namespace:
                    module = import_module(m)
                    cmd = getattr(module, c)
                    namespace[c] = staticmethod(cmd)
            except (ValueError, ImportError, AttributeError):
                raise ImportError("Command not found '{}'".format(command_fqn))

        return super(MainMeta, mcs).__new__(mcs, name, bases, namespace)


class BaseMain(metaclass=MainMeta):
    commands = []
    description = None

    def __init__(self, args=None, parse_args=True):
        self.cli = CLI()
        self.args, self.unknown_args = argparse.Namespace(), []
        if parse_args:
            self.args, self.unknown_args = self.parse_arguments(args=args)
            self.settings = self.args.settings or os.environ.get('CLINNER_SETTINGS')

            # Inject parameters related to current stage as environment variables
            self.inject()

            # Load settings
            settings.build_from_module(self.args.settings)

            # Apply quiet mode
            if self.args.quiet:
                self.cli.disable()

    def _commands_arguments(self, parser: 'argparse.ArgumentParser', parser_class=None):
        """
        Add arguments for each command to parser.

        :param parser: Parser
        """
        # Create subparser for each command
        subparsers_kwargs = {'parser_class': lambda **kwargs: parser_class(self, **kwargs)} if parser_class else {}
        subparsers = parser.add_subparsers(title='Commands', dest='command', **subparsers_kwargs)
        subparsers.required = True
        for cmd_name, cmd in command.register.items():

            subparser_opts = cmd['parser']
            if cmd['type'] == Type.SHELL:
                subparser_opts['add_help'] = False

            p = subparsers.add_parser(cmd_name, **subparser_opts)
            if callable(cmd['arguments']):
                cmd['arguments'](p)
            else:
                for argument in cmd['arguments']:
                    try:
                        if len(argument) == 2:
                            args, kwargs = argument
                        elif len(argument) == 1:
                            args = argument[0]
                            kwargs = {}
                        else:
                            args, kwargs = None, None

                        assert isinstance(args, (tuple, list))
                        assert isinstance(kwargs, dict)
                    except AssertionError:
                        raise CommandArgParseError(str(argument))
                    else:
                        p.add_argument(*args, **kwargs)

    @abstractmethod
    def add_arguments(self, parser: 'argparse.ArgumentParser'):
        """
        Add to parser all necessary arguments for this Main.

        :param parser: Argument parser.
        """
        pass

    def parse_arguments(self, args=None, parser=None, parser_class=None):
        """
        command Line application arguments.
        """
        if parser is None:
            parser = argparse.ArgumentParser(description=self.description, conflict_handler='resolve')

        # Call inner method that adds arguments from all classes (defined in metaclass)
        self._add_arguments(parser, parser_class)

        return parser.parse_known_args(args=args)

    def run_python(self, cmd, *args, **kwargs):
        """
        Run a python command in a different process.

        :param cmd: Python command.
        :param args: List of args passed to Process.
        :param kwargs: Dict of kwargs passed to Process.
        :return: Command return code.
        """
        self.cli.logger.debug('- [Python] %s.%s', str(cmd.__module__), str(cmd.__qualname__))

        result = 0

        if not getattr(self.args, 'dry_run', False):
            # Run command
            result = cmd(*args, **kwargs)

        return result

    def run_shell(self, cmd, *args, **kwargs):
        """
        Run a shell command in a different process.

        :param cmd: Shell command.
        :param args: List of args passed to Popen.
        :param kwargs: Dict of kwargs passed to Popen.
        :return: Command return code.
        """
        self.cli.logger.debug('- [Shell] %s', ' '.join(cmd))

        result = 0

        if not getattr(self.args, 'dry_run', False):
            # Run command
            p = Popen(args=cmd, *args, **kwargs)
            while p.returncode is None:  # pragma: no cover
                try:
                    p.wait()
                except KeyboardInterrupt:
                    pass

            result = p.returncode

        return result

    def run_command(self, input_command, *args, **kwargs):
        """
        Run the given command, building it with arguments.

        :param input_command: Command to execute.
        :param args: List of args passed to run_<type> command.
        :param kwargs: Dict of kwargs passed to run_<type> command.
        :return: Command return code.
        """
        # Get list of commands
        commands, command_type = Builder.build_command(input_command, *args, **kwargs)

        self.cli.logger.debug('Running commands:') if commands else None
        return_code = 0
        for c in commands:
            if command_type == Type.PYTHON:
                return_code = self.run_python(c)
            elif command_type == Type.SHELL:
                return_code = self.run_shell(c)
            else:  # pragma: no cover
                raise CommandTypeError(command_type)

            # Break on non-zero exit code.
            if return_code != 0:
                return return_code

        return return_code

    @abstractmethod
    def run(self, *args, **kwargs):
        pass



from unittest.case import TestCase
from unittest.mock import MagicMock, call

from clinner.run import DjangoCommand


class FooMain(MagicMock):
    description = 'Foo'


class FooDjangoCommand(DjangoCommand):
    main_class = FooMain


class DjangoCommandTestCase(TestCase):
    def setUp(self):
        self.command = FooDjangoCommand()

    def test_django_command_add_arguments(self):
        parser = MagicMock()
        self.command.add_arguments(parser)

        self.assertEqual(self.command.command._commands_arguments.call_count, 1)

    def test_django_command_handle(self):
        expected_calls = [call(foo='bar')]

        self.command.handle(foo='bar')

        self.assertEqual(self.command.command.run.call_args_list, expected_calls)



import shlex

from clinner.command import Type, command

__all__ = ['sphinx']


@command(command_type=Type.SHELL,
         args=((('sphinx_command',), {'help': 'Sphinx command'}),
               (('--source',), {'help': 'Sources dir', 'default': 'doc/source'}),
               (('--build',), {'help': 'Build dir', 'default': 'doc/build'})),
         parser_opts={'help': 'Sphinx doc'})
def sphinx(*args, **kwargs):
    """
    Run an sphinx command.
    """
    return [shlex.split('sphinx-build -M %s %s %s' % (kwargs['sphinx_command'], kwargs['source'], kwargs['build']))]



# -*- coding: utf-8 -*-
__version__ = '1.4.1'
__license__ = 'GPLv3'

__author__ = 'José Antonio Perdiguero López'
__email__ = 'perdy.hh@gmail.com'

__url__ = 'https://github.com/PeRDy/Clinner'
__description__ = 'command Line Interface builder that helps creating an entry point for your application'



import shlex

from clinner.command import Type, command

__all__ = ['pytest']


@command(command_type=Type.SHELL, parser_opts={'help': 'Run unit tests'})
def pytest(*args, **kwargs):
    """
    Run unit tests using pytest.
    """
    coverage_erase = shlex.split('coverage erase')

    tests = shlex.split('pytest')
    tests += args

    return [coverage_erase, tests]



import shlex
from functools import partial, update_wrapper

from typing import Callable, List, Tuple, Union

from clinner.command import Type as CommandType, command
from clinner.exceptions import CommandTypeError
from clinner.settings import settings

__all__ = ['Builder']


class Builder:
    """
    Helper for build commands stored in register.
    """

    @staticmethod
    def _build_shell_command(method, *args, **kwargs) -> List[List[str]]:
        """
        Build a shell command using given method, args and kwargs. Bash commands should return a list of bash commands
        split by shlex.

        :param method: Command callable.
        :param args: List of command args.
        :param kwargs: Dict of command kwargs.
        :return: List of commands ready to be executed.
        """
        return method(*args, **kwargs)

    @staticmethod
    def _build_python_command(method, *args, **kwargs) -> List[Callable]:
        """
        Build a python command using given method, args and kwargs. Python commands will return a python's callable
        partialized with current args and kwargs

        :param method: Command callable.
        :param args: List of command args.
        :param kwargs: Dict of command kwargs.
        :return: List of commands ready to be executed.
        """
        cmd = partial(method, *args, **kwargs)
        update_wrapper(cmd, method)
        return [cmd]

    @staticmethod
    def build_command(command_name: str, *args, **kwargs) -> Tuple[List[Union[List[str], Callable]], CommandType]:
        """
        Build command given his name and a list of args.

        :param command_name: command name.
        :param args: List of command args.
        :param kwargs: Dict of command kwargs.
        :return: List of commands ready to be executed.
        """
        # Get registered command
        cmd = command.register[command_name]

        # Get command callable
        method = cmd['callable']
        command_type = cmd['type']

        # Get default args if necessary
        if not args:
            try:
                args = shlex.split(settings.default_args[command_name])
            except KeyError:
                args = []

        if command_type == CommandType.PYTHON:
            built = Builder._build_python_command(method, *args, **kwargs)
        elif command_type == CommandType.SHELL:
            built = Builder._build_shell_command(method, *args, **kwargs)
        else:
            raise CommandTypeError(command_type)

        return built, command_type



import logging
from collections import OrderedDict

try:
    import colorlog

    _colorlog = True
except ImportError:  # pragma: no cover
    _colorlog = False

__all__ = ['CLI']


class CLI:
    """
    command Line Interface helpers, such as predefined styles for different type of messages, headers, etc.
    """
    SEP = '-' * 70

    def __init__(self):
        if _colorlog:
            self.handler = colorlog.StreamHandler()
            self.handler.setFormatter(colorlog.ColoredFormatter(
                '%(log_color)s%(message)s',
                datefmt=None,
                reset=True,
                log_colors={
                    'DEBUG': 'cyan',
                    'INFO': 'green',
                    'WARNING': 'yellow',
                    'ERROR': 'red',
                    'CRITICAL': 'red,bg_white',
                },
                style='%'
            ))
        else:
            #  Default to basic logging
            self.handler = logging.StreamHandler()

        self.logger = logging.getLogger('cli')
        self.logger.addHandler(self.handler)
        self.logger.setLevel(logging.DEBUG)

    def disable(self):
        self.logger.removeHandler(self.handler)

    def enable(self):
        self.logger.addHandler(self.handler)

    def print_return(self, code: int):
        level = logging.INFO if code == 0 else logging.ERROR
        self.logger.log(level, 'Return code: %d', code)

    def print_header(self, **kwargs):
        fields = OrderedDict(sorted(kwargs.items(), key=lambda x: x[0]))

        fmt = '{:<%d}: {}' % (max([len(x) for x in fields.keys()]),)
        header_lines = [self.SEP]
        for k, v in fields.items():
            header_lines.append(fmt.format(k.capitalize(), v))
        header_lines.append(self.SEP)

        self.logger.info('\n'.join(header_lines))



import shlex

from clinner.command import Type, command

__all__ = ['prospector']


@command(command_type=Type.SHELL,
         parser_opts={'help': 'Run prospector lint'})
def prospector(*args, **kwargs):
    """
    Run prospector lint.
    """
    return [shlex.split('prospector') + list(args)]



import logging
from unittest.case import TestCase
from unittest.mock import MagicMock, call, patch

from clinner.cli import CLI


class CLITestCase(TestCase):
    def setUp(self):
        self.cli = CLI()
        self.cli.logger = MagicMock()

    def test_init_with_colorlog(self):
        with patch('clinner.cli._colorlog', True), patch('colorlog.StreamHandler') as handler_mock:
            CLI()

            self.assertTrue(handler_mock.call_count, 1)
            self.assertTrue(handler_mock.return_value.setFormatter.call_count, 1)

    def test_init_without_colorlog(self):
        with patch('clinner.cli._colorlog', False), patch('clinner.cli.logging.StreamHandler') as handler_mock:
            CLI()

            self.assertTrue(handler_mock.call_count, 1)

    def test_disable(self):
        self.cli.disable()
        self.assertEqual(self.cli.logger.removeHandler.call_count, 1)

    def test_enable(self):
        self.cli.enable()
        self.assertEqual(self.cli.logger.addHandler.call_count, 1)

    def test_print_return_ok(self):
        self.cli.print_return(0)
        expected_calls = [call(logging.INFO, 'Return code: %d', 0)]
        self.assertCountEqual(self.cli.logger.log.call_args_list, expected_calls)

    def test_print_return_error(self):
        self.cli.print_return(1)
        expected_calls = [call(logging.ERROR, 'Return code: %d', 1)]
        self.assertCountEqual(self.cli.logger.log.call_args_list, expected_calls)

    def test_print_header(self):
        self.cli.print_header(foo=True, bar=1)
        msg = self.cli.logger.info.call_args[0][0]
        self.assertIn('Foo: True', msg)
        self.assertIn('Bar: 1', msg)

    def tearDown(self):
        pass



import argparse  # noqa

from clinner.run.base import BaseMain
from clinner.run.mixins import HealthCheckMixin

__all__ = ['Main', 'HealthCheckMain']


class Main(BaseMain):
    def add_arguments(self, parser: 'argparse.ArgumentParser'):
        """
        Add to parser all necessary arguments for this Main.

        :param parser: Argument parser.
        """
        parser.add_argument('-s', '--settings',
                            help='Module or object with Clinner settings in format "package.module[:Object]"')
        parser.add_argument('-q', '--quiet', help='Quiet mode. No standard output other than executed application',
                            action='store_true')
        parser.add_argument('--dry-run', action='store_true',
                            help='Dry run. Skip commands execution, useful to check which commands will be executed '
                                 'and execution order')

    def run(self, *args, **kwargs):
        """
        Run specified command through system arguments.

        Arguments that have been parsed properly will be passed through \**kwargs. Unknown arguments will be passed as a
        list of strings through \*args.

        This method will print a header and the return code.
        """
        cmd_args = self.unknown_args if not args else args

        cmd_kwargs = vars(self.args)
        cmd_kwargs.update(kwargs)

        command = cmd_kwargs['command']
        settings = cmd_kwargs['settings']

        self.cli.print_header(command=command, settings=settings)

        return_code = self.run_command(command, *cmd_args, **cmd_kwargs)

        self.cli.print_return(return_code)
        return return_code


class HealthCheckMain(HealthCheckMixin, Main):
    """
    Main class with health check behavior.
    """
    pass



import sys
from multiprocessing import Queue
from unittest.case import TestCase
from unittest.mock import patch

import pytest

from clinner.command import Type, command
from clinner.exceptions import CommandArgParseError, CommandTypeError, WrongCommandError
from clinner.run.main import Main


class CommandRegisterTestCase(TestCase):
    @pytest.fixture(autouse=True)
    def create_command(self):
        @command
        def foo(*args, **kwargs):
            return 42

        yield

        del command.register['foo']

    def test_command(self):
        self.assertTrue('foo' in command.register)

    def test_get_wrong_command(self):
        with self.assertRaises(WrongCommandError):
            command.register['wrong_command']


class CommandTestCase(TestCase):
    def setUp(self):
        sys.argv = ['test']
        self.cli_patcher = patch('clinner.run.base.CLI')
        self.cli_patcher.start()

    def test_command(self):
        @command
        def foo(*args, **kwargs):
            kwargs['q'].put(42)

        queue = Queue()
        args = ['foo']
        Main(args).run(q=queue)

        self.assertEqual(queue.get(), 42)

    def test_command_python(self):
        @command(command_type=Type.PYTHON)
        def foo(*args, **kwargs):
            kwargs['q'].put(42)

        queue = Queue()
        args = ['foo']
        Main(args).run(q=queue)

        self.assertEqual(queue.get(), 42)

    def test_command_with_args(self):
        @command(command_type=Type.PYTHON,
                 args=((('-b', '--bar'),),))  # args
        def foo(*args, **kwargs):
            kwargs['q'].put(kwargs['bar'])

        queue = Queue()
        args = ['foo']
        Main(args).run(bar='foobar', q=queue)

        self.assertEqual(queue.get(), 'foobar')

    def test_command_with_args_and_opts(self):
        @command(command_type=Type.PYTHON,
                 args=((('-b', '--bar'), {'type': int, 'help': 'bar argument'}),))  # args and opts
        def foo(*args, **kwargs):
            kwargs['q'].put(kwargs['bar'])

        queue = Queue()
        args = ['foo']
        Main(args).run(bar=3, q=queue)

        self.assertEqual(queue.get(), 3)

    def test_command_with_callable_args(self):
        def add_arguments(parser):
            parser.add_argument('-b', '--bar', type=int, help='bar argument')

        @command(args=add_arguments)
        def foo(*args, **kwargs):
            kwargs['q'].put(kwargs['bar'])

        queue = Queue()
        args = ['foo']
        Main(args).run(bar=3, q=queue)

        self.assertEqual(queue.get(), 3)

    def test_command_with_wrong_args_number(self):
        @command(command_type=Type.PYTHON,
                 args=(((1, 2, 3),)))  # wrong args number
        def foo(*args, **kwargs):
            kwargs['q'].put(kwargs['bar'])

        args = ['foo']
        self.assertRaises(CommandArgParseError, Main, args)

    def test_command_with_unknown_args(self):
        @command  # args
        def foo(*args, **kwargs):
            kwargs['q'].put(args[0])

        queue = Queue()
        args = ['foo']
        Main(args).run('foobar', q=queue)

        self.assertEqual(queue.get(), 'foobar')

    def test_command_shell(self):
        @command(command_type=Type.SHELL)
        def foo(*args, **kwargs):
            return [['foo']]

        args = ['foo']
        main = Main(args)
        with patch('clinner.run.base.Popen') as popen_mock:
            popen_mock.return_value.returncode = 0
            main.run()

        self.assertEqual(popen_mock.call_count, 1)
        self.assertEqual(popen_mock.call_args[1]['args'], ['foo'])

    def test_command_multiple_shell(self):
        @command(command_type=Type.SHELL)
        def foo(*args, **kwargs):
            return [['foo'], ['bar']]

        args = ['foo']
        main = Main(args)
        with patch('clinner.run.base.Popen') as popen_mock:
            popen_mock.return_value.returncode = 0
            main.run()

        self.assertEqual(popen_mock.call_count, 2)
        self.assertEqual(popen_mock.call_args_list[0][1]['args'], ['foo'])
        self.assertEqual(popen_mock.call_args_list[1][1]['args'], ['bar'])

    def test_command_multiple_shell_failing(self):
        @command(command_type=Type.SHELL)
        def foo(*args, **kwargs):
            return [['foo'], ['bar']]

        args = ['foo']
        main = Main(args)
        with patch('clinner.run.base.Popen') as popen_mock:
            popen_mock.return_value.returncode = 1
            main.run()

        self.assertEqual(popen_mock.call_count, 1)
        self.assertEqual(popen_mock.call_args_list[0][1]['args'], ['foo'])

    def test_command_wrong_type(self):
        @command(command_type='foo')
        def foo(*args, **kwargs):
            pass

        args = ['foo']
        self.assertRaises(CommandTypeError, Main(args).run, args)

    def tearDown(self):
        self.cli_patcher.stop()
        del command.register['foo']



import argparse  # noqa

try:
    from django.core.management.base import BaseCommand, CommandParser
except ImportError:
    BaseCommand = object
    CommandParser = None

__all__ = ['DjangoCommand']


class DjangoCommand(BaseCommand):
    """
    Wrapper that makes a Django command from a Main class, including parsers only for commands.
    """
    main_class = None

    def __init__(self, *args, **kwargs):
        self.help = self.main_class.description
        super(DjangoCommand, self).__init__(*args, **kwargs)
        self.command = self.main_class(parse_args=False)
        self.command.cli.disable()

    def add_arguments(self, parser):
        self.command._commands_arguments(parser=parser, parser_class=CommandParser)

    def handle(self, *args, **options):
        self.command.run(**options)



import time
from abc import ABCMeta, abstractmethod
from random import random

__all__ = ['HealthCheckMixin']


class HealthCheckMixin(metaclass=ABCMeta):
    """
    Adds health checking behavior to Main classes. To do that is necessary to define a health_check method responsible
    of return the current status of the application.

    This mixin also adds a new parameter ``-r``, ``--retry`` that defines the number of retries done after a failure.
    These retries uses an exponential backoff to calculate timing.
    """

    def add_arguments(self, parser: 'argparse.ArgumentParser'):
        parser.add_argument('-r', '--retry', help='Health check retries before run command. Disabled with 0, max 10.',
                            type=int, default=5, choices=range(11))
        parser.add_argument('--skip-check', help='Skip health check.', default=False, action='store_true')

    @abstractmethod
    def health_check(self):
        """
        Does a health check.

        :return: True if health check was successful. False otherwise.
        """
        pass

    def _health_check(self):
        """
        Does a health check and retry using exponential backoff if it fails.

        :return: True if health check was successful. False otherwise.
        """
        if not self.args.skip_check and self.args.retry:
            health = False
            self.cli.logger.info('Performing healthcheck...')
            timeout = random()

            for i in (i for i in range(self.args.retry) if not health):
                if not self.health_check():
                    self.cli.logger.warning('Health check failed, retrying ({}/{})'.format(i + 1, self.args.retry))
                    time.sleep(timeout)
                    timeout *= 2.
                else:
                    # Healthcheck successful
                    health = True

            if not health:
                self.cli.logger.error('Retry attempts exceeded, health check failed')
        else:
            health = True

        return health

    def run(self, *args, **kwargs):
        """
        Run specified command through system arguments.

        Before running the command, a health check function will be called and if result is not successful, the command
        will be aborted.

        Arguments that have been parsed properly will be passed through \**kwargs. Unknown arguments will be passed as a
        list of strings through \*args.

        This method will print a header and the return code.
        """
        cmd_args = self.unknown_args if not args else args

        cmd_kwargs = vars(self.args)
        cmd_kwargs.update(kwargs)

        command = cmd_kwargs['command']
        settings = cmd_kwargs['settings']

        self.cli.print_header(command=command, settings=settings)

        if self._health_check():
            return_code = self.run_command(command, *cmd_args, **cmd_kwargs)
        else:
            return_code = 1

        self.cli.print_return(return_code)
        return return_code



from enum import Enum
from functools import partial, update_wrapper

from typing import Any, Callable, Dict, Tuple

from clinner.exceptions import WrongCommandError

__all__ = ['command', 'Type']


class Type(Enum):
    PYTHON = 'python'
    SHELL = 'shell'
    BASH = SHELL


class CommandRegister(dict):
    """
    Register for commands.
    """

    def register(self, func: Callable, command_type: Type, arguments: Tuple[Tuple[str], Dict[str, Any]],
                 parser: Dict[str, Any]):
        self[func.__name__] = {
            'callable': func,
            'type': command_type,
            'arguments': arguments,
            'parser': parser,
        }

    def __getitem__(self, item):
        if item not in self:
            raise WrongCommandError(item)

        return dict.__getitem__(self, item)


class command:  # noqa
    """
    Decorator to register the given functions in a register, along with their command line arguments.
    """
    register = CommandRegister()

    def __init__(self, func=None, command_type=Type.PYTHON, args=None, parser_opts=None):
        """
        Decorator to register given functions in a register. This decorator allows to be used as a common decorator
        without arguments:

        @command
        def foobar(bar):
            pass

        But also is possible to provide command line arguments, as expected by argparse.ArgumentParser.add_argument:
        @command((('-f', '--foo'), {help='Foo argument that does nothing'}),  # Command argument
                 (('--bar',), {action='store_true', help='Bar argument stored as True'})  # Another command argument
                 title='foobar_command', help='Help for foobar_command')  # Parser parameters
        def foobar(*args, **kwargs):
            pass

        @command(args=((('-f', '--foo'), {'help': 'Foo argument that does nothing'}),
                       (('--bar',), {'action': 'store_true', 'help': 'Bar argument stored as True'})),
                 parser_opts={'title': 'foobar_command', 'help': 'Help for foobar_command'})
        def foobar(*args, **kwargs):
            pass

        For last, is possible to decorate functions or class methods:
        class Foo:
            @staticmethod
            @command
            def bar():
                pass

        :param func: Function or class method to be decorated.
        :param args: argparse.ArgumentParser.add_argument args.
        :param parser_opts: argparse.ArgumentParser.add_subparser kwargs.
        """
        self.args = args or ()
        self.kwargs = parser_opts or {}
        self.command_type = command_type

        if func is not None and callable(func):
            # Full initialization decorator
            self._decorate(func=func, command_type=self.command_type, args=self.args, parse_opts=self.kwargs)
        else:
            # Partial initialization decorator
            self.func = None

    def _decorate(self, func, command_type, args, parse_opts):
        self.func = func
        update_wrapper(self, func)

        self.register.register(self, command_type, args, parse_opts)

    def __get__(self, instance, owner=None):
        """
        Make it works with functions and methods.
        """
        return partial(self.__call__, instance)

    def __call__(self, *args, **kwargs):
        """
        Call function. If decorator is not fully initialized, initialize it.
        """
        if self.func:
            # Decorator behavior
            return self.func(*args, **kwargs)
        else:
            # Decorator is not initialized and now is giving the function to be decorated
            if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
                self._decorate(func=args[0], command_type=self.command_type, args=self.args, parse_opts=self.kwargs)
                return self
            else:
                raise ValueError('Decorator is not initialized')



import shlex

from clinner.command import Type, command

__all__ = ['tox']


@command(command_type=Type.SHELL,
         parser_opts={'help': 'Run tox'})
def tox(*args, **kwargs):
    """
    Run tests using tox.
    """
    return [shlex.split('tox') + list(args)]



import argparse
from unittest.case import TestCase
from unittest.mock import patch

from multiprocessing import Queue

import pytest

from clinner.command import command, Type
from clinner.run.main import Main


class BaseMainTestCase(TestCase):
    @pytest.fixture(autouse=True)
    def define_main(self):
        class FooMain(Main):
            @staticmethod
            @command
            def foo(*args, **kwargs):
                return 42

            @staticmethod
            @command(command_type=Type.SHELL)
            def bar(*args, **kwargs):
                return [[]]

            def add_arguments(self, parser: 'argparse.ArgumentParser'):
                parser.add_argument('-f', '--foo', type=int, help='')

            def inject_foo(self):
                self.foo = True

        self.main_cls = FooMain

    def setUp(self):
        self.cli_patcher = patch('clinner.run.base.CLI')
        self.cli_patcher.start()

    def test_main_inject(self):
        args = ['foo']
        main = self.main_cls(args)

        self.assertTrue(getattr(main, 'foo', False))

    def test_main_add_arguments(self):
        args = ['-f', '3', 'foo']
        main = self.main_cls(args)

        self.assertEqual(main.args.foo, 3)

    def test_main_quiet(self):
        args = ['-q', 'foo']
        main = self.main_cls(args)

        self.assertEqual(main.cli.disable.call_count, 1)

    def test_dry_run_python(self):
        args = ['--dry-run', 'foo']
        main = self.main_cls(args)

        result = main.run()

        self.assertEqual(result, 0)

    def test_dry_run_shell(self):
        args = ['--dry-run', 'bar']
        main = self.main_cls(args)

        with patch('clinner.run.base.Popen') as popen_mock:
            main.run()

        self.assertEqual(popen_mock.call_count, 0)

    def test_explicit_commands(self):
        class BarMain(Main):
            commands = ('tests.run.utils_base.foobar',)

        args = ['foobar']
        main = BarMain(args)
        queue = Queue()
        main.run(q=queue)

        self.assertEqual(queue.get(), 42)

    def test_explicit_command_wrong(self):
        with self.assertRaises(ImportError):
            class BarMain(Main):
                commands = ('tests.run.utils_base.wrong_command',)

    def test_replace_explicit_commands(self):
        class BarMain(Main):
            commands = ('tests.run.utils_base.foobar',)

            @staticmethod
            @command
            def foobar(*args, **kwargs):
                kwargs['q'].put(1)

        args = ['foobar']
        main = BarMain(args)
        queue = Queue()
        main.run(q=queue)

        self.assertEqual(queue.get(), 1)

    def test_parse_args_with_parser(self):
        parser = argparse.ArgumentParser()

        main = self.main_cls(parse_args=False)
        args, _ = main.parse_arguments(args=['-f', '1', 'foo'], parser=parser)

        self.assertEqual(args.foo, 1)

    def tearDown(self):
        self.cli_patcher.stop()
